@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2000-2020. All Rights Reserved.
+ * Copyright Ericsson AB 2000-2023. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -36,15 +36,9 @@
 #include <unistd.h>
 #include <sys/times.h>
 
-#if TIME_WITH_SYS_TIME
-# include <sys/time.h>
-# include <time.h>
-#else
-# if HAVE_SYS_TIME_H
+#include <time.h>
+#if HAVE_SYS_TIME_H
 #  include <sys/time.h>
-# else
-#  include <time.h>
-# endif
 #endif
 
 #include <sys/socket.h>
@@ -104,13 +98,13 @@ static int send_challenge(ei_cnode *ec, void *ctx, int pkt_sz,
                           unsigned challenge,
                           DistFlags version, unsigned ms);
 static int recv_challenge(ei_socket_callbacks *cbs, void *ctx, int pkt_sz,
-                          unsigned *challenge, unsigned *version,
+                          unsigned *challenge,
 			  DistFlags *flags, char *namebuf, unsigned ms);
 static int send_challenge_reply(ei_socket_callbacks *cbs, void *ctx,
                                 int pkt_sz, unsigned char digest[16], 
 				unsigned challenge, unsigned ms);
 static int recv_complement(ei_socket_callbacks *cbs, void *ctx,
-                           int pkt_sz, unsigned ms);
+                           int pkt_sz, DistFlags *flags, unsigned ms);
 static int recv_challenge_reply(ei_socket_callbacks *cbs, void *ctx,
                                 int pkt_sz, unsigned our_challenge,
 				char cookie[], 
@@ -121,11 +115,7 @@ static int send_challenge_ack(ei_socket_callbacks *cbs, void *ctx,
 static int recv_challenge_ack(ei_socket_callbacks *cbs, void *ctx, 
 			      int pkt_sz, unsigned our_challenge,
 			      char cookie[], unsigned ms);
-static int send_name(ei_cnode *ec, void *ctx, int pkt_sz,
-                     unsigned version, unsigned ms);
-static int send_complement(ei_cnode *ec, void *ctx, int pkt_sz,
-                            unsigned epmd_says_version, DistFlags her_flags,
-                            unsigned ms);
+static int send_name(ei_cnode *ec, void *ctx, int pkt_sz, unsigned ms);
 static int recv_name(ei_socket_callbacks *cbs, void *ctx, int pkt_sz,
                      char* send_name_tag, DistFlags *flags,
                      char *namebuf, unsigned ms);
@@ -143,10 +133,10 @@ dyn_gethostbyname_r(const char *name, struct hostent *hostp, char **buffer_p,
 static void abort_connection(ei_socket_callbacks *cbs, void *ctx);
 static int close_connection(ei_socket_callbacks *cbs, void *ctx, int fd);
 
-static char *
+static const char *
 estr(int e)
 {
-    char *str = strerror(e);
+    const char *str = strerror(e);
     if (!str)
         return "unknown error";
     return str;
@@ -440,7 +430,7 @@ static ei_socket_info* get_ei_socket_info(int fd)
 
 static int remove_ei_socket_info(int fd)
 {
-    return put_ei_socket_info(fd, -1, NULL, NULL, NULL, NULL);
+    return put_ei_socket_info(fd, -1, null_cookie, NULL, NULL, NULL);
 }
 
 ei_cnode *ei_fd_to_cnode(int fd)
@@ -514,11 +504,6 @@ const char *ei_thisalivename(const ei_cnode* ec)
     return ec->thisalivename;
 }
 
-short ei_thiscreation(const ei_cnode* ec)
-{
-    return ec->creation;
-}
-
 /* FIXME: this function is not an api, why not? */
 const char *ei_thiscookie(const ei_cnode* ec)
 {
@@ -624,7 +609,7 @@ int ei_make_pid(ei_cnode *ec, erlang_pid *pid)
      * of modifying the 'num' field in the pid returned by
      * ei_self(). Since 'serial' field in pid returned by
      * ei_self() is initialized to 0, pids created by
-     * ei_make_pid() wont clash with such badly created pids
+     * ei_make_pid() won't clash with such badly created pids
      * using ei_self() unless user also modified serial, but
      * that has at least never been suggested by the
      * documentation.
@@ -671,7 +656,7 @@ int ei_make_pid(ei_cnode *ec, erlang_pid *pid)
 
 #undef EI_MAKE_REF_ATOMIC__
 #ifdef _REENTRANT
-#  if ((SIZEOF_LONG == 8 || SIZEOF_LONGLONG == 8)         \
+#  if ((SIZEOF_LONG == 8 || SIZEOF_LONG_LONG == 8)        \
        && (ETHR_HAVE___atomic_compare_exchange_n & 8)     \
        && (ETHR_HAVE___atomic_load_n & 8))
 #    define EI_MAKE_REF_ATOMIC__
@@ -687,7 +672,7 @@ static ei_mutex_t *ref_mtx = NULL;
 
 /*
  * We use a global counter for all c-nodes in this process.
- * We wont wrap anyway due to the enormous amount of values
+ * We won't wrap anyway due to the enormous amount of values
  * available.
  */
 #ifdef EI_MAKE_REF_ATOMIC__
@@ -770,6 +755,8 @@ int ei_make_ref(ei_cnode *ec, erlang_ref *ref)
     ref->n[0] = ref_count[0];
     ref->n[1] = ref_count[1];
     ref->n[2] = ref_count[2];
+    ref->n[3] = 0;
+    ref->n[4] = 0;
     
     ref_count[0]++;
     ref_count[0] &= 0x3ffff;
@@ -900,13 +887,13 @@ int ei_init_connect(void)
 
 /*
 * Perhaps run this routine instead of ei_connect_init/2 ?
-* Initailize by setting:
+* Initialize by setting:
 * thishostname, thisalivename, thisnodename and thisipaddr
 */
 int ei_connect_xinit_ussi(ei_cnode* ec, const char *thishostname,
                           const char *thisalivename, const char *thisnodename,
                           Erl_IpAddr thisipaddr, const char *cookie,
-                          const short creation, ei_socket_callbacks *cbs,
+                          unsigned int creation, ei_socket_callbacks *cbs,
                           int cbs_sz, void *setup_context)
 {
     char *dbglevel;
@@ -923,6 +910,10 @@ int ei_connect_xinit_ussi(ei_cnode* ec, const char *thishostname,
     }
     
     ec->creation = creation;
+    if (ec->creation < 4) {
+        /* Avoid invalid 0-creation as well as old tiny 1,2,3 values. */
+        ec->creation += 0xE10000;
+    }
     ec->pidsn = 0;
     
     if (cookie) {
@@ -960,7 +951,7 @@ int ei_connect_xinit_ussi(ei_cnode* ec, const char *thishostname,
         strcpy(ec->self.node, thisnodename);
         ec->self.num = 0;
         ec->self.serial = 0;
-        ec->self.creation = creation;
+        ec->self.creation = ec->creation;
     }
     else {
         /* dynamic name */
@@ -985,7 +976,7 @@ int ei_connect_xinit_ussi(ei_cnode* ec, const char *thishostname,
 int ei_connect_xinit(ei_cnode* ec, const char *thishostname,
                      const char *thisalivename, const char *thisnodename,
                      Erl_IpAddr thisipaddr, const char *cookie,
-                     const short creation)
+                     unsigned int creation)
 {
     return ei_connect_xinit_ussi(ec, thishostname, thisalivename, thisnodename,
                                  thisipaddr, cookie, creation,
@@ -1000,7 +991,7 @@ int ei_connect_xinit(ei_cnode* ec, const char *thishostname,
 * otherwise return -1.
 */
 int ei_connect_init_ussi(ei_cnode* ec, const char* this_node_name,
-                         const char *cookie, short creation,
+                         const char *cookie, unsigned int creation,
                          ei_socket_callbacks *cbs, int cbs_sz,
                          void *setup_context)
 {
@@ -1027,9 +1018,7 @@ int ei_connect_init_ussi(ei_cnode* ec, const char* this_node_name,
 	return ERL_ERROR;
     }
 
-    if (this_node_name == NULL) {
-	sprintf(thisalivename, "c%d", (int) getpid());
-    } else if (strlen(this_node_name) >= sizeof(thisalivename)) {
+    if (strlen(this_node_name) >= sizeof(thisalivename)) {
 	EI_TRACE_ERR0("ei_connect_init","ERROR: this_node_name too long");
 	return ERL_ERROR;
     } else {
@@ -1069,11 +1058,11 @@ int ei_connect_init_ussi(ei_cnode* ec, const char* this_node_name,
 	    strcpy(thishostname, hp->h_name);
 	}
     }
-    if (strlen(this_node_name) + 1 + strlen(thishostname) > MAXNODELEN) {
+    if (snprintf(thisnodename, sizeof(thisnodename), "%s@%s",
+                 this_node_name, thishostname) > sizeof(thisnodename)) {
         EI_TRACE_ERR0("ei_connect_init_ussi","this node name is too long");
         return ERL_ERROR;
     }
-    sprintf(thisnodename, "%s@%s", this_node_name, thishostname);
     res = ei_connect_xinit_ussi(ec, thishostname, thisalivename, thisnodename,
                                 (struct in_addr *)*hp->h_addr_list, cookie, creation,
                                 cbs, cbs_sz, setup_context);
@@ -1083,7 +1072,7 @@ int ei_connect_init_ussi(ei_cnode* ec, const char* this_node_name,
 }
 
 int ei_connect_init(ei_cnode* ec, const char* this_node_name,
-                    const char *cookie, short creation)
+                    const char *cookie, unsigned int creation)
 {
     return ei_connect_init_ussi(ec, this_node_name, cookie, creation,
                                 &ei_default_socket_callbacks,
@@ -1146,7 +1135,7 @@ struct hostent *dyn_gethostbyname_r(const char *name,
 #endif
 }
 
-/* Finds the the IP address for hostname and saves that IP address at
+/* Finds the IP address for hostname and saves that IP address at
    the location that ip_wb points to. Returns a negative error code if
    the IP address cannot be found for the hostname. */
 static int ip_address_from_hostname(char* hostname,
@@ -1162,7 +1151,7 @@ static int ip_address_from_hostname(char* hostname,
     hp = dyn_gethostbyname_r(hostname,&host,buffer_p,buffer_size,&ei_h_errno);
     if (hp == NULL) {
 	char thishostname[EI_MAXHOSTNAMELEN+1];
-        /* gethostname requies len to be max(hostname) + 1*/
+        /* gethostname requires len to be max(hostname) + 1*/
 	if (gethostname(thishostname,EI_MAXHOSTNAMELEN+1) < 0) {
 	    EI_TRACE_ERR0("ip_address_from_hostname",
 			  "Failed to get name of this host");
@@ -1227,7 +1216,6 @@ static int ei_connect_helper(ei_cnode* ec,
     ei_socket_callbacks *cbs = ec->cbs;
     void *ctx;
     int sockd;
-    unsigned her_version;
     DistFlags her_flags;
     unsigned our_challenge, her_challenge;
     unsigned char our_digest[16];
@@ -1288,24 +1276,21 @@ static int ei_connect_helper(ei_cnode* ec,
         goto error;
     }
         
-    if (send_name(ec, ctx, pkt_sz, epmd_says_version, tmo))
+    if (send_name(ec, ctx, pkt_sz, tmo))
         goto error;
     if (recv_status(ec, ctx, pkt_sz, tmo))
         goto error;
-    if (recv_challenge(cbs, ctx, pkt_sz, &her_challenge, &her_version,
+    if (recv_challenge(cbs, ctx, pkt_sz, &her_challenge,
                        &her_flags, NULL, tmo))
         goto error;
-    her_version = (her_flags & DFLAG_HANDSHAKE_23) ? EI_DIST_6 : EI_DIST_5;
     our_challenge = gen_challenge();
     gen_digest(her_challenge, ec->ei_connect_cookie, our_digest);
-    if (send_complement(ec, ctx, pkt_sz, epmd_says_version, her_flags, tmo))
-        goto error;
     if (send_challenge_reply(cbs, ctx, pkt_sz, our_digest, our_challenge, tmo))
         goto error;
     if (recv_challenge_ack(cbs, ctx, pkt_sz, our_challenge, 
                            ec->ei_connect_cookie, tmo))
         goto error;
-    if (put_ei_socket_info(sockd, her_version, null_cookie, ec, cbs, ctx) != 0)
+    if (put_ei_socket_info(sockd, EI_DIST_6, null_cookie, ec, cbs, ctx) != 0)
         goto error;
 
     if (cbs->connect_handshake_complete) {
@@ -1630,7 +1615,6 @@ int ei_accept_tmo(ei_cnode* ec, int lfd, ErlConnect *conp, unsigned ms)
     }
     
     {
-        unsigned her_version = (her_flags & DFLAG_HANDSHAKE_23) ? 6 : 5;
 	unsigned our_challenge;
 	unsigned her_challenge;
 	unsigned char our_digest[16];
@@ -1640,8 +1624,8 @@ int ei_accept_tmo(ei_cnode* ec, int lfd, ErlConnect *conp, unsigned ms)
 	our_challenge = gen_challenge();
 	if (send_challenge(ec, ctx, pkt_sz, our_challenge, her_flags, tmo))
 	    goto error;
-        if (send_name_tag == 'n' && (her_flags & DFLAG_HANDSHAKE_23)) {
-            if (recv_complement(cbs, ctx, pkt_sz, tmo))
+        if (send_name_tag == 'n') {
+            if (recv_complement(cbs, ctx, pkt_sz, &her_flags, tmo))
                 goto error;
         }
 	if (recv_challenge_reply(cbs, ctx, pkt_sz, our_challenge, 
@@ -1650,7 +1634,7 @@ int ei_accept_tmo(ei_cnode* ec, int lfd, ErlConnect *conp, unsigned ms)
 	gen_digest(her_challenge, ec->ei_connect_cookie, our_digest);
 	if (send_challenge_ack(cbs, ctx, pkt_sz, our_digest, tmo))
 	    goto error;
-	if (put_ei_socket_info(fd, her_version, null_cookie, ec, cbs, ctx) != 0)
+        if (put_ei_socket_info(fd, EI_DIST_6, null_cookie, ec, cbs, ctx) != 0)
             goto error;
     }
     if (conp) {
@@ -1845,18 +1829,26 @@ int ei_xreceive_msg_tmo(int fd, erlang_msg *msg, ei_x_buff *x, unsigned ms)
     return ei_do_receive_msg(fd, 0, msg, x, ms);
 }
 
-/* 
-* The RPC consists of two parts, send and receive.
-* Here is the send part ! 
-* { PidFrom, { call, Mod, Fun, Args, user }} 
-*/
 /*
-* Now returns non-negative number for success, negative for failure.
+* A remote process call consists of two parts, sending a request and
+* receiving a response. This function sends the request and the
+* ei_rpc_from function receives the response.
+*
+* Here is the term that is sent when (flags & EI_RPC_FETCH_STDOUT) != 0:
+*
+* { PidFrom, { call, Mod, Fun, Args, send_stdout_to_caller }}
+*
+* Here is the term that is sent otherwise:
+*
+* { PidFrom, { call, Mod, Fun, Args, user }}
+*
+* Returns a non-negative number for success and a negative number for
+* failure.
+*
 */
-int ei_rpc_to(ei_cnode *ec, int fd, char *mod, char *fun,
-	      const char *buf, int len)
+int ei_xrpc_to(ei_cnode *ec, int fd, char *mod, char *fun,
+               const char *buf, int len, int flags)
 {
-
     ei_x_buff x;
     erlang_pid *self = ei_self(ec);
     int err = ERL_ERROR;
@@ -1866,10 +1858,10 @@ int ei_rpc_to(ei_cnode *ec, int fd, char *mod, char *fun,
         goto einval;
     if (ei_x_encode_tuple_header(&x, 2) < 0)  /* A */
         goto einval;
-    
+
     if (ei_x_encode_pid(&x, self) < 0)	      /* A 1 */
         goto einval;
-    
+
     if (ei_x_encode_tuple_header(&x, 5) < 0)  /* B A 2 */
         goto einval;
     if (ei_x_encode_atom(&x, "call") < 0)     /* B 1 */
@@ -1880,14 +1872,19 @@ int ei_rpc_to(ei_cnode *ec, int fd, char *mod, char *fun,
         goto einval;
     if (ei_x_append_buf(&x, buf, len) < 0)    /* B 4 */
         goto einval;
-    if (ei_x_encode_atom(&x, "user") < 0)     /* B 5 */
-        goto einval;
-    
+    if (flags & EI_RPC_FETCH_STDOUT) {
+        if (ei_x_encode_atom(&x, "send_stdout_to_caller") < 0)     /* B 5 */
+            goto einval;
+    } else {
+        if (ei_x_encode_atom(&x, "user") < 0)     /* B 5 */
+            goto einval;
+    }
+
     err = ei_send_reg_encoded(fd, self, "rex", x.buff, x.index);
     if (err)
         goto error;
-    
-    ei_x_free(&x);	
+
+    ei_x_free(&x);
 
     return 0;
 
@@ -1898,6 +1895,13 @@ error:
     if (x.buff != NULL)
         ei_x_free(&x);
     return err;
+} /* xrpc_to */
+
+
+int ei_rpc_to(ei_cnode *ec, int fd, char *mod, char *fun,
+	      const char *buf, int len)
+{
+    return ei_xrpc_to(ec, fd, mod, fun, buf, len, 0);
 } /* rpc_to */
 
   /*
@@ -2019,10 +2023,10 @@ static unsigned int gen_challenge(void)
     uname(&s.name);
     s.cpu  = clock();
     s.pid  = getpid();
-#ifndef __ANDROID__
-    s.hid  = gethostid();
-#else
+#if defined(__ANDROID__) || defined(__HAIKU__)
     s.hid  = 0;
+#else
+    s.hid  = gethostid();
 #endif
     s.uid  = getuid();
     s.gid  = getgid();
@@ -2131,7 +2135,7 @@ static int send_status(ei_socket_callbacks *cbs, void *ctx,
     char *buf, *s;
     char dbuf[DEFBUF_SIZ];
     int siz = strlen(status) + 1 + pkt_sz;
-    int err;
+    int err, ret;
     ssize_t len;
 
     buf = (siz > DEFBUF_SIZ) ? malloc(siz) : dbuf;
@@ -2148,7 +2152,8 @@ static int send_status(ei_socket_callbacks *cbs, void *ctx,
         put32be(s,siz - 4);
         break;
     default:
-        return -1;
+        ret = -1;
+        goto done;
     }
     put8(s, 's');
     memcpy(s, status, strlen(status));
@@ -2159,16 +2164,17 @@ static int send_status(ei_socket_callbacks *cbs, void *ctx,
     if (err) {
 	EI_TRACE_ERR2("send_status","-> SEND_STATUS socket write failed: %s (%d)",
                       estr(err), err);
-	if (buf != dbuf)
-	    free(buf);        
         EI_CONN_SAVE_ERRNO__(err);
-	return -1;
+	ret = -1;
     }
-    EI_TRACE_CONN1("send_status","-> SEND_STATUS (%s)",status);
-
+    else {
+        EI_TRACE_CONN1("send_status","-> SEND_STATUS (%s)",status);
+        ret =  0;
+    }
+done:
     if (buf != dbuf)
 	free(buf);
-    return 0;
+    return ret;
 }
 
 static int recv_status(ei_cnode *ec, void *ctx,
@@ -2242,30 +2248,16 @@ error:
 static DistFlags preferred_flags(void)
 {
     DistFlags flags =
-        DFLAG_EXTENDED_REFERENCES
+        DFLAG_MANDATORY_25_DIGEST
+        | DFLAG_DIST_MANDATORY
         | DFLAG_DIST_MONITOR
-        | DFLAG_EXTENDED_PIDS_PORTS
-        | DFLAG_FUN_TAGS
-        | DFLAG_NEW_FUN_TAGS
-        | DFLAG_NEW_FLOATS
-        | DFLAG_SMALL_ATOM_TAGS
-        | DFLAG_UTF8_ATOMS
-        | DFLAG_MAP_TAG
-        | DFLAG_BIG_CREATION
-        | DFLAG_EXPORT_PTR_TAG
-        | DFLAG_BIT_BINARIES
-        | DFLAG_HANDSHAKE_23;
-    if (ei_internal_use_21_bitstr_expfun()) {
-        flags &= ~(DFLAG_EXPORT_PTR_TAG
-                   | DFLAG_BIT_BINARIES);
-    }
+        | DFLAG_SMALL_ATOM_TAGS;
     return flags;
 }
 
 static int send_name(ei_cnode *ec,
                      void *ctx,
                      int pkt_sz,
-                     unsigned version,
                      unsigned ms)
 {
     char *buf;
@@ -2274,28 +2266,21 @@ static int send_name(ei_cnode *ec,
     const char* name_ptr;
     unsigned int name_len;
     int siz;
-    int err;
+    int err, ret;
     ssize_t len;
     DistFlags flags = preferred_flags();
-    char tag;
 
     if (ec->thisnodename[0]) {
         name_ptr = ec->thisnodename;
-        tag = (version == EI_DIST_5) ? 'n' : 'N';
     }
     else {
         /* dynamic node name */
         name_ptr = ec->thishostname;
-        tag = 'N'; /* presume ver 6 */
         flags |= DFLAG_NAME_ME;
     }
 
     name_len = strlen(name_ptr);
-
-    if (tag == 'n')
-        siz = pkt_sz + 1 + 2 + 4 + name_len;
-    else
-        siz = pkt_sz + 1 + 8 + 4 + 2 + name_len;
+    siz = pkt_sz + 1 + 8 + 4 + 2 + name_len;
 
     buf = (siz > DEFBUF_SIZ) ? malloc(siz) : dbuf;
     if (!buf) {
@@ -2311,19 +2296,14 @@ static int send_name(ei_cnode *ec,
         put32be(s,siz - 4);
         break;
     default:
-        return -1;
+        ret = -1;
+        goto done;
     }
 
-    put8(s, tag);
-    if (tag == 'n') {
-        put16be(s, EI_DIST_5); /* some impl (jinterface) demand ver==5 */
-        put32be(s, flags);
-    }
-    else { /* tag == 'N' */
-        put64be(s, flags);
-        put32be(s, ec->creation);
-        put16be(s, name_len);
-    }
+    put8(s, 'N');
+    put64be(s, flags);
+    put32be(s, ec->creation);
+    put16be(s, name_len);
     memcpy(s, name_ptr, name_len);
     len = (ssize_t) siz;
     err = ei_write_fill_ctx_t__(ec->cbs, ctx, buf, &len, ms);
@@ -2331,15 +2311,16 @@ static int send_name(ei_cnode *ec,
         err = EIO;
     if (err) {
 	EI_TRACE_ERR0("send_name", "SEND_NAME -> socket write failed");
-	if (buf != dbuf)
-	    free(buf);
         EI_CONN_SAVE_ERRNO__(err);
-	return -1;
+	ret = -1;
     }
-
+    else {
+        ret = 0;
+    }
+done:
     if (buf != dbuf)
 	free(buf);
-    return 0;
+    return ret;
 }
 
 static int send_challenge(ei_cnode *ec,
@@ -2354,15 +2335,11 @@ static int send_challenge(ei_cnode *ec,
     char dbuf[DEFBUF_SIZ];
     const unsigned int nodename_len = strlen(ec->thisnodename);
     int siz;
-    int err;
+    int err, ret;
     ssize_t len;
     DistFlags flags;
-    const char tag = (her_flags & DFLAG_HANDSHAKE_23) ? 'N' : 'n';
 
-    if (tag == 'n')
-        siz = pkt_sz + 1 + 2 + 4 + 4 + nodename_len;
-    else
-        siz = pkt_sz + 1 + 8 + 4 + 4 + 2 + nodename_len;
+    siz = pkt_sz + 1 + 8 + 4 + 4 + 2 + nodename_len;
 
     buf = (siz > DEFBUF_SIZ) ? malloc(siz) : dbuf;
     if (!buf) {
@@ -2378,22 +2355,16 @@ static int send_challenge(ei_cnode *ec,
         put32be(s,siz - 4);
         break;
     default:
-        return -1;
+        ret = -1;
+        goto done;
     }
 
     flags = preferred_flags();
-    put8(s, tag);
-    if (tag == 'n') {
-        put16be(s, EI_DIST_5);  /* choosen version */
-        put32be(s, flags);
-        put32be(s, challenge);
-    }
-    else {
-        put64be(s, flags);
-        put32be(s, challenge);
-        put32be(s, ec->creation);
-        put16be(s, nodename_len);
-    }
+    put8(s, 'N');
+    put64be(s, flags);
+    put32be(s, challenge);
+    put32be(s, ec->creation);
+    put16be(s, nodename_len);
     memcpy(s, ec->thisnodename, nodename_len);
     len = (ssize_t) siz;
     err = ei_write_fill_ctx_t__(ec->cbs, ctx, buf, &len, ms);
@@ -2401,19 +2372,19 @@ static int send_challenge(ei_cnode *ec,
         err = EIO;
     if (err) {
 	EI_TRACE_ERR0("send_challenge", "-> SEND_CHALLENGE socket write failed");
-	if (buf != dbuf)
-	    free(buf);
         EI_CONN_SAVE_ERRNO__(err);
-	return -1;
+	ret = -1;
     }
-    
+    else
+        ret = 0;
+done:
     if (buf != dbuf)
 	free(buf);
-    return 0;
+    return ret;
 }
 
 static int recv_challenge(ei_socket_callbacks *cbs, void *ctx,
-                          int pkt_sz, unsigned *challenge, unsigned *version,
+                          int pkt_sz, unsigned *challenge,
 			  DistFlags *flags, char *namebuf, unsigned ms)
 {
     char dbuf[DEFBUF_SIZ];
@@ -2435,48 +2406,26 @@ static int recv_challenge(ei_socket_callbacks *cbs, void *ctx,
     }
     s = buf;
     tag = get8(s);
-    if (tag != 'n' && tag != 'N') {
+    if (tag != 'N') {
 	EI_TRACE_ERR2("recv_challenge",
 		      "<- RECV_CHALLENGE incorrect tag, "
-		      "expected 'n' or 'N', got '%c' (%u)",tag,tag);
+		      "expected 'N', got '%c' (%u)", tag, tag);
 	goto error;
     }
-    if (tag == 'n') { /* OLD */
-        unsigned int version;
-        if (rlen < 1+2+4+4) {
-            EI_TRACE_ERR1("recv_challenge","<- RECV_CHALLENGE 'n' packet too short (%d)",
-                          rlen)
-            goto error;
-        }
-
-        version = get16be(s);
-        if (version != EI_DIST_5) {
-            EI_TRACE_ERR1("recv_challenge",
-                          "<- RECV_CHALLENGE 'n' incorrect version=%d",
-                          version);
-            goto error;
-        }
-        *flags = get32be(s);
-        *challenge = get32be(s);
-        nodename_len = (buf + rlen) - s;
+    if (rlen < 1+8+4+4+2) {
+        EI_TRACE_ERR1("recv_challenge","<- RECV_CHALLENGE 'N' packet too short (%d)",
+                      rlen)
+        goto error;
     }
-    else { /* NEW */
-        if (rlen < 1+8+4+4+2) {
-            EI_TRACE_ERR1("recv_challenge","<- RECV_CHALLENGE 'N' packet too short (%d)",
-                          rlen)
-            goto error;
-        }
-        *version = EI_DIST_6;
-        *flags = get64be(s);
-        *challenge = get32be(s);
-        s += 4; /* ignore peer 'creation' */
-        nodename_len = get16be(s);
-        if (nodename_len > (buf + rlen) - s) {
-            EI_TRACE_ERR1("recv_challenge",
-                          "<- RECV_CHALLENGE 'N' nodename too long (%d)",
-                          nodename_len);
-            goto error;
-        }
+    *flags = get64be(s);
+    *challenge = get32be(s);
+    s += 4; /* ignore peer 'creation' */
+    nodename_len = get16be(s);
+    if (nodename_len > (buf + rlen) - s) {
+        EI_TRACE_ERR1("recv_challenge",
+                      "<- RECV_CHALLENGE 'N' nodename too long (%d)",
+                      nodename_len);
+        goto error;
     }
 
     if (nodename_len > MAXNODELEN) {
@@ -2485,22 +2434,13 @@ static int recv_challenge(ei_socket_callbacks *cbs, void *ctx,
         goto error;
     }
 
-    if (!(*flags & DFLAG_EXTENDED_REFERENCES)) {
-	EI_TRACE_ERR0("recv_challenge","<- RECV_CHALLENGE peer cannot "
-		      "handle extended references");
-	goto error;
+    if (*flags & DFLAG_MANDATORY_25_DIGEST) {
+        *flags |= DFLAG_DIST_MANDATORY_25;
     }
 
-    if (!(*flags & DFLAG_EXTENDED_PIDS_PORTS)) {
+    if ((*flags & DFLAG_DIST_MANDATORY) != DFLAG_DIST_MANDATORY) {
 	EI_TRACE_ERR0("recv_challenge","<- RECV_CHALLENGE peer cannot "
-		      "handle extended pids and ports");
-	erl_errno = EIO;
-	goto error;
-    }
-	    
-    if (!(*flags & DFLAG_NEW_FLOATS)) {
-	EI_TRACE_ERR0("recv_challenge","<- RECV_CHALLENGE peer cannot "
-		      "handle binary float encoding");
+		      "handle all mandatory capabilities");
 	goto error;
     }
 
@@ -2512,12 +2452,10 @@ static int recv_challenge(ei_socket_callbacks *cbs, void *ctx,
     
     if (!is_static)
 	free(buf);
-    EI_TRACE_CONN4("recv_challenge","<- RECV_CHALLENGE (ok) node = %s, "
-	    "version = %u, "
+    EI_TRACE_CONN3("recv_challenge","<- RECV_CHALLENGE (ok) node = %s, "
 	    "flags = %u, "
 	    "challenge = %d",
 	    namebuf,
-	    *version,
 	    *flags,
 	    *challenge
 	    );
@@ -2529,63 +2467,6 @@ error:
     return -1;
 }
 
-static int send_complement(ei_cnode *ec,
-                            void *ctx,
-                            int pkt_sz,
-                            unsigned epmd_says_version,
-                            DistFlags her_flags,
-                            unsigned ms)
-{
-    if (epmd_says_version == EI_DIST_5 && (her_flags & DFLAG_HANDSHAKE_23)) {
-        char *buf;
-        unsigned char *s;
-        char dbuf[DEFBUF_SIZ];
-        int err;
-        ssize_t len;
-        unsigned int flagsHigh;
-        const int siz = pkt_sz + 1 + 4 + 4;
-
-        buf = (siz > DEFBUF_SIZ) ? malloc(siz) : dbuf;
-        if (!buf) {
-            erl_errno = ENOMEM;
-            return -1;
-        }
-        s = (unsigned char *)buf;
-        switch (pkt_sz) {
-        case 2:
-            put16be(s,siz - 2);
-            break;
-        case 4:
-            put32be(s,siz - 4);
-            break;
-        default:
-            return -1;
-        }
-        flagsHigh = preferred_flags() >> 32;
-
-        put8(s, 'c');
-        put32be(s, flagsHigh);
-        put32be(s, ec->creation);
-
-        len = (ssize_t) siz;
-        err = ei_write_fill_ctx_t__(ec->cbs, ctx, buf, &len, ms);
-        if (!err && len != (ssize_t) siz)
-            err = EIO;
-        if (err) {
-            EI_TRACE_ERR0("send_name", "SEND_NAME -> socket write failed");
-            if (buf != dbuf)
-                free(buf);
-            EI_CONN_SAVE_ERRNO__(err);
-            return -1;
-        }
-
-        if (buf != dbuf)
-            free(buf);
-    }
-    return 0;
-}
-
-
 static int send_challenge_reply(ei_socket_callbacks *cbs, void *ctx,
                                 int pkt_sz, unsigned char digest[16], 
 				unsigned challenge, unsigned ms) 
@@ -2635,6 +2516,7 @@ static int send_challenge_reply(ei_socket_callbacks *cbs, void *ctx,
 static int recv_complement(ei_socket_callbacks *cbs,
                            void *ctx,
                            int pkt_sz,
+                           DistFlags *flags,
                            unsigned ms)
 {
     char dbuf[DEFBUF_SIZ];
@@ -2661,6 +2543,14 @@ static int recv_complement(ei_socket_callbacks *cbs,
 		      "expected 'c' got '%c' (%u)",tag,tag);
 	goto error;
     }
+    *flags |= (DistFlags)get32be(s) << 32;
+
+    if ((~*flags) & (DFLAG_DIST_MANDATORY | DFLAG_HANDSHAKE_23)) {
+	EI_TRACE_ERR0("recv_complement","<- RECV_COMPLEMENT peer cannot "
+		      "handle all mandatory capabilities");
+	goto error;
+    }
+
     creation = get32be(s);
     if (!is_static)
 	free(buf);
@@ -2848,6 +2738,7 @@ static int recv_name(ei_socket_callbacks *cbs, void *ctx,
     char *s;
     char tmp_nodename[MAXNODELEN+1];
     char tag;
+    DistFlags flag_mask;
     
     erl_errno = EIO;		/* Default */
 
@@ -2878,6 +2769,7 @@ static int recv_name(ei_socket_callbacks *cbs, void *ctx,
             goto error;
         }
         *flags = get32be(s);
+        flag_mask = ((DistFlags)1 << 32) - 1;
         namelen = rlen - (1+2+4);
     }
     else { /* tag == 'N' */
@@ -2887,19 +2779,18 @@ static int recv_name(ei_socket_callbacks *cbs, void *ctx,
             goto error;
         }
         *flags = get64be(s);
+        flag_mask = ~(DistFlags)0;
         s += 4; /* ignore peer 'creation' */
         namelen = get16be(s);
     }
 
-    if (!(*flags & DFLAG_EXTENDED_REFERENCES)) {
-	EI_TRACE_ERR0("recv_name","<- RECV_NAME peer cannot handle"
-		      "extended references");
-	goto error;
+    if (*flags & DFLAG_MANDATORY_25_DIGEST) {
+        *flags |= DFLAG_DIST_MANDATORY_25;
     }
 
-    if (!(*flags & DFLAG_EXTENDED_PIDS_PORTS)) {
+    if ((~*flags) & flag_mask & (DFLAG_DIST_MANDATORY | DFLAG_HANDSHAKE_23)) {
 	EI_TRACE_ERR0("recv_name","<- RECV_NAME peer cannot "
-		      "handle extended pids and ports");
+		      "handle all mandatory capabilities");
 	erl_errno = EIO;
 	goto error;
     }
