@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1998-2022. All Rights Reserved.
+%% Copyright Ericsson AB 1998-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -29,7 +29,9 @@
          otp_8562/1, otp_8665/1, otp_8911/1, otp_10302/1, otp_10820/1,
          otp_11728/1, encoding/1, extends/1,  function_macro/1,
 	 test_error/1, test_warning/1, otp_14285/1,
-	 test_if/1,source_name/1,otp_16978/1,otp_16824/1,scan_file/1,file_macro/1]).
+	 test_if/1,source_name/1,otp_16978/1,otp_16824/1,scan_file/1,file_macro/1,
+         triple_quotes_warning/1,
+   deterministic_include/1, nondeterministic_include/1]).
 
 -export([epp_parse_erl_form/2]).
 
@@ -50,6 +52,7 @@ config(data_dir, _) ->
     filename:absname("./epp_SUITE_data").
 -else.
 -include_lib("common_test/include/ct.hrl").
+-include_lib("stdlib/include/assert.hrl").
 -export([init_per_testcase/2, end_per_testcase/2]).
 
 init_per_testcase(_, Config) ->
@@ -70,7 +73,9 @@ all() ->
      overload_mac, otp_8388, otp_8470, otp_8562,
      otp_8665, otp_8911, otp_10302, otp_10820, otp_11728,
      encoding, extends, function_macro, test_error, test_warning,
-     otp_14285, test_if, source_name, otp_16978, otp_16824, scan_file, file_macro].
+     otp_14285, test_if, source_name, otp_16978, otp_16824, scan_file, file_macro,
+     triple_quotes_warning,
+     deterministic_include, nondeterministic_include].
 
 groups() ->
     [{upcase_mac, [], [upcase_mac_1, upcase_mac_2]},
@@ -124,6 +129,64 @@ file_macro(Config) when is_list(Config) ->
     "Other source" = FileA = FileB,
     ok.
 
+deterministic_include(Config) when is_list(Config) ->
+    DataDir = proplists:get_value(data_dir, Config),
+    File = filename:join(DataDir, "deterministic_include.erl"),
+    {ok, List} = epp:parse_file(File, [{includes, [DataDir]},
+                                       {deterministic, true},
+                                       {source_name, "deterministic_include.erl"}]),
+
+    %% In deterministic mode, only basenames, rather than full paths, should
+    %% be written to the -file() attributes resulting from -include and -include_lib
+    ?assert(lists:any(fun
+                       ({attribute,_Anno,file,{"baz.hrl",_Line}}) -> true;
+                       (_) -> false
+                     end,
+                     List),
+            "Expected a basename in the -file attribute resulting from " ++
+            "including baz.hrl in deterministic mode."),
+    ?assert(lists:any(fun
+                       ({attribute,_Anno,file,{"file.hrl",_Line}}) -> true;
+                       (_) -> false
+                     end,
+                     List),
+            "Expected a basename in the -file attribute resulting from " ++
+            "including file.hrl in deterministic mode."),
+    ok.
+
+nondeterministic_include(Config) when is_list(Config) ->
+    DataDir = proplists:get_value(data_dir, Config),
+    File = filename:join(DataDir, "deterministic_include.erl"),
+    {ok, List} = epp:parse_file(File, [{includes, [DataDir]},
+                                       {source_name, "deterministic_include.erl"}]),
+
+    %% Outside of deterministic mode, full paths, should be written to
+    %% the -file() attributes resulting from -include and -include_lib
+    %% to make debugging easier.
+    %% We don't try to assume what the full absolute path will be in the
+    %% unit test, since that can depend on the environment and how the
+    %% test is executed. Instead, we just look for whether there is
+    %% the parent directory along with the basename at least.
+    IncludeAbsolutePathSuffix = filename:join("include","baz.hrl"),
+    ?assert(lists:any(fun
+                       ({attribute,_Anno,file,{IncludePath,_Line}}) ->
+                         lists:suffix(IncludeAbsolutePathSuffix,IncludePath);
+                       (_) -> false
+                     end,
+                     List),
+            "Expected an absolute in the -file attribute resulting from " ++
+            "including baz.hrl outside of deterministic mode."),
+    IncludeLibAbsolutePathSuffix = filename:join("include","file.hrl"),
+    ?assert(lists:any(fun
+                       ({attribute,_Anno,file,{IncludePath,_line}}) ->
+                         lists:suffix(IncludeLibAbsolutePathSuffix,IncludePath);
+                       (_) -> false
+                     end,
+                     List),
+            "Expected an absolute in the -file attribute resulting from " ++
+            "including file.hrl outside of deterministic mode."),
+    ok.
+
 %%% Here is a little reimplementation of epp:parse_file, which times out
 %%% after 4 seconds if the epp server doesn't respond. If we use the
 %%% regular epp:parse_file, the test case will time out, and then epp
@@ -1994,6 +2057,73 @@ otp_16824(Config) when is_list(Config) ->
     [] = compile(Config, Cs),
     ok.
 
+triple_quotes_warning(Config) when is_list(Config) ->
+    Cs1 =
+        [{triple_quotes_warning_1,
+          <<"\n-doc \"foo\".\n">>,
+          []},
+         {triple_quotes_warning_2,
+          <<"\n-doc \"\" \"foo\" \"\".\n">>,
+          []},
+         {triple_quotes_warning_3,
+          <<"\n"
+            "-doc \"\"\"\n"
+            "    foo\n"
+            "    \"\"\".\n">>,
+          {warnings,[{{2,6},epp,tqstring}]}},
+         {triple_quotes_warning_4,
+          <<"\n"
+            "-doc  \"\"\"\"\n"
+            "      \"\"\"\".\n">>,
+          {warnings,
+           [{{2,7},epp,tqstring},
+            {{3,7},epp,tqstring}]}},
+         {triple_quotes_warning_5,
+          <<"\n"
+            "-doc   \"\"\"\"\"\n"
+            "       foo\n"
+            "       \"\"\"\"\".\n">>,
+          {warnings,
+           [{{2,8},epp,tqstring},
+            {{4,9},epp,tqstring}]}}
+        ],
+    [] = compile(Config, Cs1),
+
+    Cs2 =
+        [{triple_quotes_warning_10,
+          <<"\n"
+            "-export([foo/0]).\n"
+            "foo() ->\n"
+            "    \"\"\"\n"
+            "    bar\n"
+            "    \"\"\".\n">>,
+          {warnings,[{{4,5},epp,tqstring}]}},
+         {triple_quotes_warning_11,
+          <<"\n"
+            "-export([foo/0]).\n"
+            "foo() ->\n"
+            "    \"\"\"\"\n"
+            "    ++ lists:duplicate(4, $x) ++\n"
+            "    \"\"\"\".\n">>,
+          {warnings,
+           [{{4,5},epp,tqstring},
+            {{6,5},epp,tqstring}]}},
+         {triple_quotes_warning_12,
+          <<"\n"
+            "-export([foo/0]).\n"
+            "foo() ->\n"
+            "    \"\"\"\"\"\n"
+            "    bar\n"
+            "    \"\"\"\"\".\n">>,
+          {warnings,
+           [{{4,5},epp,tqstring},
+            {{6,6},epp,tqstring}]}} ],
+    [] = compile(Config, Cs2),
+
+    ok.
+
+
+
 %% Start location is 1.
 check(Config, Tests) ->
     eval_tests(Config, fun check_test/3, Tests).
@@ -2024,6 +2154,8 @@ eval_tests(Config, Fun, Tests) ->
                         case E of
                             {errors, Errors} ->
                                 call_format_error(Errors);
+                            {warnings, Errors} ->
+                                call_format_error(Errors);
                             _ ->
                                 ok
                         end,
