@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2020. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -22,7 +22,7 @@
 -export([all/0, suite/0,groups/0,init_per_suite/1, end_per_suite/1, 
 	 init_per_group/2,end_per_group/2, pcre/1,compile_options/1,
 	 run_options/1,combined_options/1,replace_autogen/1,
-	 global_capture/1,replace_input_types/1,replace_return/1,
+	 global_capture/1,replace_input_types/1,replace_with_fun/1,replace_return/1,
 	 split_autogen/1,split_options/1,split_specials/1,
 	 error_handling/1,pcre_cve_2008_2371/1,re_version/1,
 	 pcre_compile_workspace_overflow/1,re_infinite_loop/1, 
@@ -42,7 +42,7 @@ suite() ->
 all() -> 
     [pcre, compile_options, run_options, combined_options,
      replace_autogen, global_capture, replace_input_types,
-     replace_return, split_autogen, split_options,
+     replace_with_fun, replace_return, split_autogen, split_options,
      split_specials, error_handling, pcre_cve_2008_2371,
      pcre_compile_workspace_overflow, re_infinite_loop, 
      re_backwards_accented, opt_dupnames, opt_all_names, 
@@ -162,7 +162,7 @@ run_options(Config) when is_list(Config) ->
     {match,["ABCabcdABC","abcd"]} = re:run("ABCabcdABC",MP,[{capture,all,list}]),
     {match,[<<"ABCabcdABC">>,<<"abcd">>]} = re:run("ABCabcdABC",MP,[{capture,all,binary}]),
     {match,[{0,10}]} = re:run("ABCabcdABC",MP,[{capture,first}]),
-    {match,[{0,10}]} = re:run("ABCabcdABC",MP,[{capture,first,index}]),       ?line {match,["ABCabcdABC"]} = re:run("ABCabcdABC",MP,[{capture,first,list}]),
+    {match,[{0,10}]} = re:run("ABCabcdABC",MP,[{capture,first,index}]),       {match,["ABCabcdABC"]} = re:run("ABCabcdABC",MP,[{capture,first,list}]),
     {match,[<<"ABCabcdABC">>]} = re:run("ABCabcdABC",MP,[{capture,first,binary}]),
 
     {match,[{3,4}]} = re:run("ABCabcdABC",MP,[{capture,all_but_first}]),
@@ -365,6 +365,16 @@ replace_input_types(Config) when is_list(Config) ->
     <<"a",208,128,"cd">> = re:replace(<<"abcd">>,"b","\x{400}",[{return,binary},unicode]),
     ok.
 
+%% Test replace with a replacement function.
+replace_with_fun(Config) when is_list(Config) ->
+    <<"ABCD">> = re:replace("abcd", ".", fun(<<C>>, []) -> <<(C - $a + $A)>> end, [global, {return, binary}]),
+    <<"AbCd">> = re:replace("abcd", ".", fun(<<C>>, []) when (C - $a) rem 2 =:= 0 -> <<(C - $a + $A)>>; (C, []) -> C end, [global, {return, binary}]),
+    <<"b-ad-c">> = re:replace("abcd", "(.)(.)", fun(_, [A, B]) -> <<B/binary, $-, A/binary>> end, [global, {return, binary}]),
+    <<"#ab-B#cd">> = re:replace("abcd", ".(.)", fun(Whole, [<<C>>]) -> <<$#, Whole/binary, $-, (C - $a + $A), $#>> end, [{return, binary}]),
+    <<"#ab#cd">> = re:replace("abcd", ".(x)?(.)", fun(Whole, [<<>>, _]) -> <<$#, Whole/binary, $#>> end, [{return, binary}]),
+    <<"#ab#cd">> = re:replace("abcd", ".(.)(x)?", fun(Whole, [_]) -> <<$#, Whole/binary, $#>> end, [{return, binary}]),
+    ok.
+
 %% Test return options of replace together with global searching.
 replace_return(Config) when is_list(Config) ->
     {'EXIT',{badarg,_}} = (catch re:replace("na","(a","")),
@@ -673,7 +683,7 @@ pcre_compile_workspace_overflow(Config) when is_list(Config) ->
         {error, {"parentheses are too deeply nested (stack check)" = Str, _No}} ->
             {comment, ExpStr ++ Str};
         Other ->
-            ?t:fail({unexpected, Other})
+            ct:fail({unexpected, Other})
     end.
 
 %% Make sure matches that really loop infinitely actually fail.
@@ -1003,13 +1013,16 @@ bad_utf8_subject(Config) when is_list(Config) ->
 
 error_info(_Config) ->
     BadRegexp = {re_pattern,0,0,0,<<"xyz">>},
+    BadErr = "neither an iodata term",
     {ok,GoodRegexp} = re:compile(".*"),
+    InvalidRegexp = <<"(.*))">>,
+    InvalidErr = "could not parse regular expression\n.*unmatched parentheses.*",
 
     L = [{compile, [not_iodata]},
-         {compile, [not_iodata, not_list]},
+         {compile, [not_iodata, not_list],[{1,".*"},{2,".*"}]},
          {compile, [<<".*">>, [a|b]]},
          {compile, [<<".*">>, [bad_option]]},
-         {compile, [{a,b}, [bad_option]]},
+         {compile, [{a,b}, [bad_option]],[{1,".*"},{2,".*"}]},
 
          {grun, 3},                             %Internal.
 
@@ -1020,29 +1033,36 @@ error_info(_Config) ->
 
          {internal_run, 4},                     %Internal.
 
-         {replace, [{a,b}, {x,y}, {z,z}]},
-         {replace, [{a,b}, BadRegexp, {z,z}]},
+         {replace, [{a,b}, {x,y}, {z,z}],[{1,".*"},{2,".*"},{3,".*"}]},
+         {replace, [{a,b}, BadRegexp, {z,z}],[{1,".*"},{2,BadErr},{3,".*"}]},
+         {replace, [{a,b}, InvalidRegexp, {z,z}],[{1,".*"},{2,InvalidErr},{3,".*"}]},
 
-         {replace, [{a,b}, {x,y}, {z,z}, [a|b]]},
-         {replace, [{a,b}, BadRegexp, [bad_option]]},
-         {replace, ["", "", {z,z}, not_a_list]},
+         {replace, [{a,b}, {x,y}, {z,z}, [a|b]],[{1,".*"},{2,".*"},{3,".*"},{4,".*"}]},
+         {replace, [{a,b}, BadRegexp, [bad_option]],[{1,".*"},{2,BadErr},{3,".*"}]},
+         {replace, [{a,b}, InvalidRegexp, [bad_option]],[{1,".*"},{2,InvalidErr},{3,".*"}]},
+         {replace, ["", "", {z,z}, not_a_list],[{3,".*"},{4,".*"}]},
 
-         {run, [{a,b}, {x,y}]},
+         {run, [{a,b}, {x,y}],[{1,".*"},{2,".*"}]},
          {run, [{a,b}, ".*"]},
          {run, ["abc", {x,y}]},
-         {run, ["abc", BadRegexp]},
+         {run, ["abc", BadRegexp],[{2,BadErr}]},
+         {run, ["abc", InvalidRegexp],[{2,InvalidErr}]},
 
-         {run, [{a,b}, {x,y}, []]},
-         {run, ["abc", BadRegexp, []]},
-         {run, [{a,b}, {x,y}, [a|b]]},
-         {run, [{a,b}, ".*", bad_options]},
-         {run, ["abc", {x,y}, [bad_option]]},
-         {run, ["abc", BadRegexp, 9999]},
+         {run, [{a,b}, {x,y}, []],[{1,".*"},{2,".*"}]},
+         {run, ["abc", BadRegexp, []],[{2,BadErr}]},
+         {run, ["abc", InvalidRegexp, []],[{2,InvalidErr}]},
+         {run, [{a,b}, {x,y}, [a|b]],[{1,".*"},{2,".*"},{3,".*"}]},
+         {run, [{a,b}, ".*", bad_options],[{1,".*"},{3,".*"}]},
+         {run, ["abc", {x,y}, [bad_option]],[{2,".*"},{3,".*"}]},
+         {run, ["abc", BadRegexp, 9999],[{2,BadErr},{3,".*"}]},
+         {run, ["abc", InvalidRegexp, 9999],[{2,InvalidErr},{3,".*"}]},
 
-         {split, ["abc", BadRegexp]},
+         {split, ["abc", BadRegexp],[{2,BadErr}]},
+         {split, ["abc", InvalidRegexp],[{2,InvalidErr}]},
          {split, [{a,b}, ".*"]},
 
-         {split, ["abc", BadRegexp, [a|b]]},
+         {split, ["abc", BadRegexp, [a|b]],[{2,BadErr},{3,".*"}]},
+         {split, ["abc", InvalidRegexp, [a|b]],[{2,InvalidErr},{3,".*"}]},
          {split, [{a,b}, ".*", [bad_option]]},
 
          {ucompile, 2},                         %Internal.
